package es.uca.xmleye.gui.html;

/*
 Copyright (C) 2005-2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import java.awt.Color;
import java.awt.KeyboardFocusManager;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.JEditorPane;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.JViewport;
import javax.swing.KeyStroke;
import javax.swing.SwingConstants;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.HyperlinkEvent;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultHighlighter;
import javax.swing.text.Highlighter;
import javax.swing.text.html.HTML;
import javax.swing.text.html.HTMLDocument;

/**
 * <p>
 * JEditorPane especializado para navegación por código HTML.
 * </p>
 * 
 * <p>
 * Incorpora algunos cambios sobre el comportamiento por defecto de un
 * JEditorPane que mejoran su usabilidad, y el acceso por teclado.
 * </p>
 * 
 * <p>
 * Se supone que el documento que se tiene abierto es un HTMLDocument.
 * </p>
 * 
 * <p>
 * Siguiendo las recomendaciones de <a
 * href="http://java.sun.com/j2se/1.5.0/docs/api/java/awt/doc-files/FocusSpec.html#KeyEventDispatcher">aquí</a>,
 * para salir de este componente se debe usar CTRL-TAB o CTRL-SHIFT-TAB. Ello
 * nos permite usar TAB y SHIFT-TAB para navegar por los enlaces.
 * </p>
 */
public class XHTMLViewerPane extends JEditorPane implements DocumentListener {
	/**
	 * 
	 */
	private static final long serialVersionUID = -2949287167947038981L;
	private static final String MSG_DESPLAZAR_ABAJO = "msg-scroll-down";
	private static final String MSG_DESPLAZAR_ARRIBA = "msg-scroll-up";
	private static final String MSG_DESPLAZAR_IZQ = "msg-scroll-left";
	private static final String MSG_DESPLAZAR_DER = "msg-scroll-right";
	private static final String MSG_DESPLAZAR_AVPAG = "msg-scroll-pagina-down";
	private static final String MSG_DESPLAZAR_REPAG = "msg-scroll-pagina-up";
	private static final String MSG_DESPLAZAR_INICIO = "msg-scroll-home";
	private static final String MSG_DESPLAZAR_FIN = "msg-scroll-end";
	private static final String MSG_CYCLE_LINK_FORWARDS = "msg-cycle-link-fwd";
	private static final String MSG_CYCLE_LINK_BACKWARDS = "msg-cycle-link-bck";
	private static final String MSG_LANZAR_ENLACE = "msg-launch-link";

	// Dirección del enlace marcado actualmente, si lo hay
	private String hrefLinkSeleccionado = null;
	// Referencia a la marca con la que se destaca el enlace actual
	private Object marcaActual = null;
	// Dibujador de la marca del enlace actual
	private static Highlighter.HighlightPainter DIBUJADOR = new DefaultHighlighter.DefaultHighlightPainter(
			Color.MAGENTA);

	// Listas de enlaces, con inicios y finales en el texto
	private ArrayList<String> listaHref = new ArrayList<String>();
	private ArrayList<Integer> listaInicio = new ArrayList<Integer>();
	private ArrayList<Integer> listaFin = new ArrayList<Integer>();

	/* -------------------- CONSTRUCTORES -------------------- */

	public XHTMLViewerPane() {
		super();
		inicObservadores();
		inicEventosTecla();
	}

	public XHTMLViewerPane(String url) throws IOException {
		super(url);
		inicObservadores();
		inicEventosTecla();
	}

	public XHTMLViewerPane(String tipo, String texto) {
		super(tipo, texto);
		inicObservadores();
		inicEventosTecla();
	}

	public XHTMLViewerPane(URL pagInicial) throws IOException {
		super(pagInicial);
		inicObservadores();
		inicEventosTecla();
	}

	/* -------------------- Interfaz DocumentListener ---------- */

	public void changedUpdate(DocumentEvent e) {
		resetLinks();
	}

	public void insertUpdate(DocumentEvent e) {
	}

	public void removeUpdate(DocumentEvent e) {
	}

	/* -------------------- MÉTODOS PÚBLICOS -------------------- */

	/**
	 * <p>
	 * Instala el <code>XHTMLViewerPane</code> dentro del JScrollPane
	 * <code>pane</code>.
	 * </p>
	 * 
	 * <p>
	 * Sitúa el XHTMLViewerPane en el interior de <code>pane</code>, y además
	 * modifica las acciones por teclado para que el desplazamiento vertical
	 * funcione como en un navegador, y no como en un editor de texto: es decir,
	 * independientemente de la posición del cursor de inserción.
	 * </p>
	 * 
	 * @param pane
	 *            JScrollPane en el que situar este control.
	 */
	public void instalar(final JScrollPane pane) {
		pane.setViewportView(this);

		// Mapa de teclas a nombres de acción
		getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0),
				MSG_DESPLAZAR_ARRIBA);
		getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0),
				MSG_DESPLAZAR_ABAJO);
		getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, 0),
				MSG_DESPLAZAR_IZQ);
		getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, 0),
				MSG_DESPLAZAR_DER);
		getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_HOME, 0),
				MSG_DESPLAZAR_INICIO);
		getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_END, 0),
				MSG_DESPLAZAR_FIN);
		getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP, 0),
				MSG_DESPLAZAR_REPAG);
		getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN, 0),
				MSG_DESPLAZAR_AVPAG);

		// Mapa de nombres de acción a acciones
		getActionMap().put(
				MSG_DESPLAZAR_ARRIBA,
				new AccionDesplazar(this, pane, SwingConstants.VERTICAL, -1,
						false));
		getActionMap().put(
				MSG_DESPLAZAR_ABAJO,
				new AccionDesplazar(this, pane, SwingConstants.VERTICAL, 1,
						false));
		getActionMap().put(
				MSG_DESPLAZAR_IZQ,
				new AccionDesplazar(this, pane, SwingConstants.HORIZONTAL, -1,
						false));
		getActionMap().put(
				MSG_DESPLAZAR_DER,
				new AccionDesplazar(this, pane, SwingConstants.HORIZONTAL, 1,
						false));
		getActionMap().put(
				MSG_DESPLAZAR_REPAG,
				new AccionDesplazar(this, pane, SwingConstants.VERTICAL, -1,
						true));
		getActionMap().put(
				MSG_DESPLAZAR_AVPAG,
				new AccionDesplazar(this, pane, SwingConstants.VERTICAL, 1,
						true));

		final JScrollBar barraVert = pane.getVerticalScrollBar();
		getActionMap().put(MSG_DESPLAZAR_INICIO, new AbstractAction() {
			/**
			 * 
			 */
			private static final long serialVersionUID = 2945574825116199490L;

			public void actionPerformed(ActionEvent e) {
				barraVert.setValue(barraVert.getMinimum());
			}
		});
		getActionMap().put(MSG_DESPLAZAR_FIN, new AbstractAction() {
			/**
			 * 
			 */
			private static final long serialVersionUID = -5366373354506768668L;

			public void actionPerformed(ActionEvent e) {
				barraVert.setValue(barraVert.getMaximum());
			}
		});
	}

	@Override
	public void setCaretPosition(int p) {
		super.setCaretPosition(p);
		if (marcaActual != null) {
			getHighlighter().removeHighlight(marcaActual);
		}
	}

	/* -------------------- MÉTODOS PRIVADOS -------------------- */

	/**
	 * Registra al XHTMLViewerPane como observador del documento.
	 */
	private void inicObservadores() {
		getDocument().addDocumentListener(this);

		addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				setCaretPosition(viewToModel(e.getPoint()));
			}
		});
	}

	/**
	 * Registra las teclas de recorrido y lanzamiento de enlaces, y cambia las
	 * teclas de cambio de foco.
	 */
	private void inicEventosTecla() {

		// Foco hacia delante: CTRL-TAB
		final Set<KeyStroke> fwdFocusKeys = new HashSet<KeyStroke>();
		fwdFocusKeys.add(KeyStroke.getKeyStroke(KeyEvent.VK_TAB,
				InputEvent.CTRL_MASK));
		setFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS,
				fwdFocusKeys);

		// Foco hacia atrás: CTRL-SHIFT-TAB
		final Set<KeyStroke> bckFocusKeys = new HashSet<KeyStroke>();
		bckFocusKeys.add(KeyStroke.getKeyStroke(KeyEvent.VK_TAB,
				InputEvent.CTRL_MASK | InputEvent.SHIFT_MASK));
		setFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS,
				bckFocusKeys);

		getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_TAB, 0),
				MSG_CYCLE_LINK_FORWARDS);
		getInputMap().put(
				KeyStroke.getKeyStroke(KeyEvent.VK_TAB, InputEvent.SHIFT_MASK),
				MSG_CYCLE_LINK_BACKWARDS);
		getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0),
				MSG_LANZAR_ENLACE);

		getActionMap().put(MSG_CYCLE_LINK_FORWARDS, new AccionSiguienteLink());
		getActionMap().put(MSG_CYCLE_LINK_BACKWARDS, new AccionAnteriorLink());
		getActionMap().put(MSG_LANZAR_ENLACE, new AccionLanzarLink());
	}

	/**
	 * Registra un cambio en la dirección del enlace seleccionado por el
	 * usuario.
	 * 
	 * @param nuevaHref
	 *            Nueva dirección del enlace.
	 */
	private void setHrefLinkSeleccionado(String nuevaHref) {
		this.hrefLinkSeleccionado = nuevaHref;
	}

	/**
	 * Devuelve la dirección del enlace seleccionado actualmente por el usuario.
	 * 
	 * @return Dirección del último enlace seleccionado por el usuario, o
	 *         <code>null</code> si no hay ninguno seleccionado actualmente.
	 */
	private String getHrefLinkSeleccionado() {
		return hrefLinkSeleccionado;
	}

	/**
	 * Destaca y prepara para lanzamiento el enlace número <code>i</code>.
	 */
	private void marcarLink(int i) {
		try {
			final int inicio = listaInicio.get(i);
			final int fin = listaFin.get(i);
			final String href = listaHref.get(i);

			setHrefLinkSeleccionado(href);
			setCaretPosition(inicio);
			marcaActual = getHighlighter().addHighlight(inicio, fin, DIBUJADOR);
		} catch (BadLocationException ex) {
		}
	}

	/**
	 * Regenera la lista de enlaces y borra todas las marcas.
	 */
	private void resetLinks() {
		final HTMLDocument doc = (HTMLDocument) getDocument();

		marcaActual = null;
		setHrefLinkSeleccionado(null);

		// Rehacemos la lista de enlaces
		listaHref.clear();
		listaInicio.clear();
		listaFin.clear();
		setCaretPosition(0);

		for (HTMLDocument.Iterator iter = doc.getIterator(HTML.Tag.A); iter
				.isValid(); iter.next()) {
			listaHref.add((String) iter.getAttributes().getAttribute(
					HTML.Attribute.HREF));
			listaInicio.add(iter.getStartOffset());
			listaFin.add(iter.getEndOffset());
		}
	}

	/* -------------------- CLASES INTERNAS -------------------- */

	/**
	 * Acción a usar para desplazamientos unidad. Siempre causa un
	 * desplazamiento al lanzarse, en la dirección que se le especifique.
	 */
	private static class AccionDesplazar extends AbstractAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = -1388490750976460664L;
		private JEditorPane editor;
		private JViewport vista;
		private JScrollBar barra;
		private int orientacion;
		private int direccion;
		private boolean pagina;

		/**
		 * Crea una nueva acción a usar para desplazamientos.
		 * 
		 * @param editor
		 *            JEditorPane cuyo contenido hay que desplazar.
		 * @param scroll
		 *            JScrollPane cuyas barras hay que controlar.
		 * @param orientacion
		 *            Orientación: <code>SwingConstants.HORIZONTAL</code> o
		 *            <code>SwingConstants.VERTICAL</code>.
		 * @param direccion
		 *            Menor que 0 para arriba/izquierda, mayor para
		 *            abajo/derecha.
		 * @param pagina
		 *            <code>true</code> realiza los desplazamientos a nivel de
		 *            página, <code>false</code> hace que se realicen a nivel
		 *            de línea.
		 */
		public AccionDesplazar(JEditorPane editor, JScrollPane scroll,
				int orientacion, int direccion, boolean pagina) {
			this.editor = editor;
			this.orientacion = orientacion;
			this.direccion = (direccion > 0 ? 1 : -1);

			this.vista = scroll.getViewport();
			this.barra = (orientacion == SwingConstants.HORIZONTAL ? scroll
					.getHorizontalScrollBar() : scroll.getVerticalScrollBar());
			this.pagina = pagina;
		}

		public void actionPerformed(ActionEvent e) {
			if (barra != null) {
				Rectangle rectVista = vista.getViewRect();
				int delta = (pagina ? editor.getScrollableBlockIncrement(
						rectVista, orientacion, direccion) : editor
						.getScrollableUnitIncrement(rectVista, orientacion,
								direccion));
				barra.setValue(Math.min(barra.getValue() + direccion * delta,
						barra.getMaximum()));
			}
		}
	}

	/**
	 * <p>
	 * Acción que recorre de forma cíclica hacia delante los enlaces del
	 * documento HTML que contiene un JEditorPane.
	 * </p>
	 */
	private class AccionSiguienteLink extends AbstractAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = -4359362688267206811L;

		public void actionPerformed(ActionEvent e) {
			if (listaHref.size() > 0) {

				final int posCursor = getCaretPosition();

				// Una nota: si no existe nodo tal que posCursor < iniciar,
				// entonces posCursor se halla tras todo enlace, y el
				// siguiente es el primero, completando el ciclo).
				int coincidencia = 0;

				// Buscamos el primer enlace desde la posición del cursor en
				// la dirección dada (es decir, el primero tal que posCursor <
				// iniciar)
				for (int i = 0; i < listaHref.size(); ++i) {
					if (posCursor < listaInicio.get(i)) {
						coincidencia = i;
						break;
					}
				}

				marcarLink(coincidencia);
			}
		}
	}

	/**
	 * <p>
	 * Acción que recorre de forma cíclica hacia atrás los enlaces del documento
	 * HTML que contiene un JEditorPane.
	 * </p>
	 */
	private class AccionAnteriorLink extends AbstractAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = -3093360612169035956L;

		public void actionPerformed(ActionEvent e) {
			if (listaHref.size() > 0) {

				final int posCursor = getCaretPosition();

				// Una nota: si no existe nodo tal que posCursor > end,
				// entonces posCursor se halla antes de todo enlace, y el
				// siguiente es el último, completando el ciclo).
				int coincidencia = listaHref.size() - 1;

				// Buscamos el primer enlace desde la posición del cursor en
				// la dirección dada (es decir, el primero tal que posCursor >
				// end)
				for (int i = listaHref.size() - 1; i >= 0; --i) {
					if (posCursor > listaFin.get(i)) {
						coincidencia = i;
						break;
					}
				}

				marcarLink(coincidencia);
			}
		}
	}

	/**
	 * Lanza el enlace seleccionado actualmente por el usuario, si hay alguno. A
	 * veces dichos enlaces no tendrán una URL bien formada: sólo incluiremos la
	 * descripción entonces.
	 */
	private class AccionLanzarLink extends AbstractAction {

		/**
		 * 
		 */
		private static final long serialVersionUID = -5901606982431531852L;

		public void actionPerformed(ActionEvent e) {
			URL url;
			try {
				url = new URL(getPage(), getHrefLinkSeleccionado());
			} catch (MalformedURLException ex) {
				url = null;
			}

			fireHyperlinkUpdate(new HyperlinkEvent(XHTMLViewerPane.this,
					HyperlinkEvent.EventType.ACTIVATED, url,
					getHrefLinkSeleccionado()));
		}
	}

}
