package es.uca.xmleye.gui.trees;

/*
  Copyright (C) 2005-2008 Antonio Garcia Dominguez

  This program is free software; you can redistribute it
  and/or modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation;
  either version 2 of the License, or (at your option) any
  later version.

  This program is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE. See the GNU General Public License for more
  details.

  You should have received a copy of the GNU General Public
  License along with this program; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA 02111-1307 USA
*/

import java.io.*;
import java.util.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import javax.xml.parsers.*;

import org.w3c.dom.*;
import org.xml.sax.SAXException;
import javax.xml.xpath.*;

import es.uca.xmleye.ParsingException;
import es.uca.xmleye.SearchException;
import es.uca.xmleye.xpath.XPathPathManager;

/**
 * TreeModel basado en el API DOM de la W3C.
 *
 * @see es.uca.xmleye.gui.trees.DOMTreeModelTest
 */
public class DOMTreeModel implements TreeModel {
  private Document docXML;
  private ArrayList<TreeModelListener> obsArbol = new ArrayList<TreeModelListener>();

  /** *** CONSTRUCTORES ***************************** */

  /**
   * Crea un documento vacío. Útil para crear nuestros propios árboles usando
   * el API DOM.
   */
  public DOMTreeModel() throws ParsingException {
    try {
      DocumentBuilder docB = DocumentBuilderFactory.newInstance()
          .newDocumentBuilder();
      docXML = docB.newDocument();
    } catch (ParserConfigurationException pEx) {
      throw new ParsingException(pEx);
    }
  }

  /**
   * Usa un árbol DOM XML ya existente.
   *
   * @param d
   *            Árbol DOM de un documento XML ya analizado.
   */
  public DOMTreeModel(Document d) {
    docXML = d;
  }

  /**
   * Usa un flujo abierto por el usuario con un documento XML.
   *
   * @param is
   *            Flujo con fichero XML a analizar.
   */
  public DOMTreeModel(InputStream is) throws ParsingException {
    try {
      DocumentBuilder dBuilder = DocumentBuilderFactory.newInstance()
          .newDocumentBuilder();
      docXML = dBuilder.parse(is);
    } catch (ParserConfigurationException pEx) {
      throw new ParsingException(pEx);
    } catch (SAXException saxEx) {
      throw new ParsingException(saxEx);
    } catch (IOException ioEx) {
      throw new ParsingException(ioEx);
    }
  }

  /** *** METODOS PROPIOS *************************** */

  /**
   * Accesor al árbol DOM original.
   *
   * @return Árbol DOM usado por el modelo.
   */
  public Document getDocumento() {
    return docXML;
  }

  /**
   * Devuelve el camino al nodo padre del nodo con camino <code>camNodo</code>,
   * o <code>null</code> si no existe dicho nodo, o si la ruta no es válida.
   *
   * Suponemos que los elementos de <code>camNodo</code> pertenecen a este
   * árbol.
   */
  public TreePath getPathToParent(TreePath camNodo) {
    synchronized(docXML) {
      if (camNodo == null || camNodo.getPathCount() <= 1) {
        return null;
      } else {
        return camNodo.getParentPath();
      }
    }
  }

  /**
   * Devuelve el camino al primer hijo del nodo con camino
   * <code>camNodo</code>, o <code>null</code> si no existe dicho nodo, o
   * si la ruta no es válida.
   *
   * Suponemos que los elementos de <code>camNodo</code> pertenecen a este
   * árbol.
   */
  public TreePath getPathToFirstChild(TreePath camNodo) {
    synchronized(docXML) {
      if (camNodo == null)
        return null;

      // Debe de haber hijos
      final Object padre = camNodo.getLastPathComponent();
      if (getChildCount(padre) == 0) {
        return null;
      } else {
        return camNodo.pathByAddingChild(getChild(padre, 0));
      }
    }
  }

  /**
   *
   * Devuelve el camino al siguiente hermano del nodo con camino
   * <code>camNodo</code>, o <code>null</code> si no existe dicho nodo, o
   * si la ruta no es válida.
   *
   * Suponemos que los elementos de <code>camNodo</code> pertenecen a este
   * árbol.
   */
  public TreePath getPathToNextSibling(TreePath camNodo) {
    return getPathToSibling(camNodo, 1);
  }

  /**
   * Devuelve el camino al anterior hermano del nodo con camino
   * <code>camNodo</code>, o <code>null</code> si no existe dicho nodo, o
   * si la ruta no es válida.
   *
   * Suponemos que los elementos de <code>camNodo</code> pertenecen a este
   * árbol.
   */
  public TreePath getPathToPreviousSibling(TreePath camNodo) {
    return getPathToSibling(camNodo, -1);
  }

  /**
   * Devuelve el camino de nodos necesario para llegar a un nodo identificado
   * por una ruta XPath.
   *
   * @param sXPath
   *            Cadena con la ruta XPath que identifica al nodo.
   * @return TreePath con todos los nodos en el camino desde la raiz hasta el
   *         nodo especificado, o <code>null</code> si no es posible.
   */
  public TreePath getPathToNode(String sXPath) throws SearchException {
    synchronized(docXML) {
      try {
        final Node raiz = (Node) getRoot();
        final Node origen = XPathPathManager.getNode(sXPath, raiz)
            .item(0);
        if (origen == null)
          return null;

        return nodePath(origen);
      } catch (XPathExpressionException xpathEx) {
        throw new SearchException(xpathEx);
      }
    }
  }

  /**
   * Devuelve el camino de nodos necesario para llegar a un nodo.
   *
   * @param destino
   *            Nodo destino del camino.
   * @return TreePath con todos los nodos en el camino desde la raiz hasta
   *         <code>destino</code>, o <code>null</code> si
   *         <code>destino</code> no pertenece al árbol.
   */
  public TreePath nodePath(Node destino) {
    synchronized(docXML) {
      final Node raiz = (Node) getRoot();

      // Vamos hacia arriba hasta la raiz
      final LinkedList<Node> pilaPath = new LinkedList<Node>();
      {
        Node nodo = destino;
        while (nodo != null && !nodo.equals(raiz)) {
          pilaPath.addLast(nodo);
          nodo = nodo.getParentNode();
        }

        // El nodo no se halla en el árbol
        if (nodo == null) {
          return null;
        }
      }
      pilaPath.addLast(raiz);

      // Rehacemos el camino, añadiendo a un TreePath
      TreePath selArbol = new TreePath(pilaPath.removeLast());
      while (!pilaPath.isEmpty()) {
        selArbol = selArbol.pathByAddingChild(pilaPath.removeLast());
      }

      return selArbol;
    }
  }

  /* ---------- IMPLEMENTACION DE INTERFAZ TreeModel ---------- */

  public Object getChild(Object padre, int indice) {
    synchronized(docXML) {
      final NodeList nodes = ((Node) padre).getChildNodes();

      {
        int indiceElem = 0;
        for (int i = 0; i < nodes.getLength(); ++i) {
          final Node item = nodes.item(i);
          if (filterNode(item)) {
            if (indiceElem == indice)
              return item;
            else
              ++indiceElem;
          }
        }
      }

      return null;
    }
  }

  public int getChildCount(Object padre) {
    return countChildren(padre);
  }

  public int getIndexOfChild(Object padre, Object child) {
    synchronized(docXML) {
      if (padre == null || child == null)
        return -1;

      final Node nodoPadre = (Node) padre;
      final Node nodoBuscado = (Node) child;

      // Comprobamos que pertenecen a nuestro recorrido
      if (!filterNode(nodoPadre) || !filterNode(nodoBuscado))
        return -1;

      // Buscamos entre los elementos hijos de padre
      final NodeList children = nodoPadre.getChildNodes();
      int elementIndex = 0;

      for (int i = 0; i < children.getLength(); ++i) {
        final Node nodoHijo = children.item(i);

        if (nodoBuscado.equals(nodoHijo)) {
          return elementIndex;
        } else if (filterNode(nodoHijo)) {
          ++elementIndex;
        }
      }

      // Nodo no hallado.
      return -1;
    }
  }

  /**
   * Devuelve la raíz del árbol. Se trata del nodo principal de tipo elemento
   * (y no el nodo que representa al documento completo, padre del primero).
   *
   * @return Raíz del árbol DOM.
   */
  public Object getRoot() {
    synchronized(docXML) {
      return docXML.getDocumentElement();
    }
  }

  public boolean isLeaf(Object nodo) {
    return countChildren(nodo) == 0;
  }

  public void valueForPathChanged(TreePath ruta, Object nuevoValor) {
  }

  public void addTreeModelListener(TreeModelListener l) {
    synchronized(obsArbol) {
      obsArbol.add(l);
    }
  }

  public void removeTreeModelListener(TreeModelListener l) {
    synchronized(obsArbol) {
      obsArbol.remove(l);
    }
  }

  /** * METODOS PROTEGIDOS ************************** */

  /**
   * Este método sirve para distinguir entre los nodos que formarán parte de
   * la vista y cuáles no, a pesar de que pertenezcan al árbol DOM XML
   * original.
   */
  protected boolean filterNode(Node n) {
    synchronized(docXML) {
      return n.getNodeType() == Node.ELEMENT_NODE;
    }
  }

  /** * METODOS PRIVADOS **************************** */

  /**
   * <p>
   * Counts how many children <code>parent</code> has.
   * </p>
   *
   * <p> This method encapsulates some common functionality
   * among several public methods, helping make them independent
   * and less prone to mistakes while redefining them in a
   * subclass.</p>
   *
   * @see getChildCount
   * @see esHoja
   */
  private int countChildren(Object parent) {
    synchronized(docXML) {
      if (parent instanceof Element) {
        synchronized(parent) {
          NodeList nodes = ((Element) parent).getChildNodes();
          int count = 0;

          synchronized(nodes) {
            for (int i = 0; i < nodes.getLength(); ++i) {
              if (filterNode(nodes.item(i))) {
                ++count;
              }
            }
          }

          return count;
        }
      } else {
        return 0;
      }
    }
  }

  /**
   * Método genérico para manipulación de caminos con hermanos, yendo hacia
   * atrás o hacia delante. Devuelve el camino al hermano del nodo con camino
   * <code>camNodo</code> con desplazamiento <code>delta</code>, donde
   * valores positivos indican los hermanos siguientes y negativos los
   * anteriores.
   *
   * Se devuelve <code>null</code> si la ruta no es válida o el nodo
   * solicitado no existe.
   *
   * Se supone que los elementos de <code>camNodo</code> son de este árbol.
   */
  private TreePath getPathToSibling(TreePath camNodo, int delta) {
    synchronized(docXML) {
      // Buscamos al padre
      final TreePath camPadre = getPathToParent(camNodo);
      if (camPadre == null)
        return null;

      final Object padre = camPadre.getLastPathComponent();
      final Object actual = camNodo.getLastPathComponent();

      final int numHijos = getChildCount(padre);
      final int indiceActual = getIndexOfChild(padre, actual);

      if (!siblingCondition(delta, indiceActual, numHijos)) {
        return null;
      }
      final Object hermano = getChild(padre,
                                      siblingPosition(delta,
                                                      indiceActual));

      return camPadre.pathByAddingChild(hermano);
    }
  }

  /**
   * Devuelve verdadero si se puede pasar al hermano en la direccion dada.
   *
   * @param direccion
   *            -1 si vamos al anterior, 1 si vamos al siguiente.
   * @param indice
   *            Indice del nodo actual.
   * @param cantidad
   *            Numero de hijos del padre comun.
   * @return <code>true</code> si existe un hermano en la dirección que
   *         queremos, <code>false</code> de lo contrario.
   */
  private static boolean siblingCondition(int direccion,
                                          int indice,
                                          int cantidad) {
    switch (direccion) {
      case -1:
        return (indice > 0 && indice <= cantidad - 1);
      case 1:
        return (indice >= 0 && indice < cantidad - 1);
      default:
        return false;
    }
  }

  /**
   * Devuelve el índice del hermano en cuestion.
   *
   * @param direccion
   *            -1 si vamos al anterior, 1 si vamos al siguiente.
   * @param indice
   *            Indice del nodo actual.
   * @return Índice del hermano en la dirección dada.
   */
  private static int siblingPosition(int direccion,
                                     int indice) {
    switch (direccion) {
      case 1:
      case -1:
        return indice + direccion;
      default:
        return indice;
    }
  }

}
