package es.uca.xmleye.xpath;

/*
 Copyright (C) 2005-2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import java.util.*;
import javax.xml.xpath.*;

/**
 * <p>
 * Extensión XPath <code>wholeWordMatch(pajar, aguja)</code> para localizar
 * palabras completas.
 * </p>
 * 
 * <p>
 * También sirve para cadenas que no sean palabras (:INDUCTION, por ejemplo): en
 * dicho caso busca espacios o límites de línea.
 * </p>
 */
public class WholeWordMatchXPathFunction implements XPathFunction {
	/** Nombre de la función. */
	public final static String NOMBRE_LOCAL = "wholeWordMatch";
	/** Aridad de la función. */
	public final static int ARIDAD = 2;

	/**
	 * Busca en el texto una ocurrencia completa de la palabra o cadena
	 * especificada.
	 * 
	 * @param args
	 *            Lista de argumentos: texto y clave de búsqueda.
	 * @return <code>Boolean</code> con valor <code>true</code> si hay
	 *         coincidencia, valor <code>false</code> si no.
	 */
	@SuppressWarnings("unchecked")
	public Object evaluate(List args) {
		if (args.size() != ARIDAD)
			return null;
		final String cadena = (String) args.get(0);
		String clave = (String) args.get(1);

		if (clave.length() == 0)
			return true;

		// Ponemos los delimitadores necesarios
		final String delimInicio = Character.isLetterOrDigit(clave.charAt(0)) ? "\\b"
				: "(?:\\s|^)";
		final String delimFin = Character.isLetterOrDigit(clave.charAt(clave
				.length() - 1)) ? "\\b" : "(?:\\s|$)";

		// Escapamos la clave de búsqueda
		clave = clave.replaceAll("([{}\\[\\]()^$.|*+?\\\\])", "\\\\$1");

		final String regexp = "(?s).*" + delimInicio + clave + delimFin + ".*";
		final boolean bCoincidencia = cadena.matches(regexp);

		return new Boolean(bCoincidencia);
	}

	/**
	 * Devuelve el FQN, Fully Qualified Name.
	 * 
	 * @return Nombre completamente calificado en el formato
	 *         namespace:parte_local.
	 */
	public static String getFQN() {
		return CustomXPathFunctionResolver.XPATH_EXT_NS_PREFIX + ":" + NOMBRE_LOCAL;
	}
}
