package es.uca.xmleye.formats;

/*
  Copyright (C) 2005-2008 Antonio García Domínguez

  This program is free software; you can redistribute it
  and/or modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation;
  either version 2 of the License, or (at your option) any
  later version.

  This program is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE. See the GNU General Public License for more
  details.

  You should have received a copy of the GNU General Public
  License along with this program; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA 02111-1307 USA
*/

import junit.framework.*;
import junit.textui.*;

import java.io.File;
import java.util.*;

import es.uca.xmleye.formats.FormatDescriptor;
import es.uca.xmleye.formats.FormatDescriptorException;
import es.uca.xmleye.formats.FormatDescriptorRepository;

public class FormatDescriptorRepositoryTest extends TestCase {
  final private static String REPO_DIR     = "build/formats";
  final private static String REPO_DIR_ALT = "build/formats3";
  final private static String TEST_FORMAT    = "xml.format";

  private FormatDescriptorRepository repositorio;

  /** Crea un nuevo caso de prueba. */
  public FormatDescriptorRepositoryTest(String nombre) {
    super(nombre);
  }

  /** Crea el conjunto de pruebas para JUnit. */
  public static Test suite() {
    return new TestSuite(FormatDescriptorRepositoryTest.class);
  }

  /** Lanza el caso de prueba mediante JUnit. */
  public static void main(String[] args) {
    TestRunner.run(suite());
  }

  /** Probamos que el repositorio se abre correctamente. */
  public void testApertura() {
    assertEquals("El repositorio tiene que tener una sola entrada",
                 1, repositorio.getSize());

    Locale.setDefault(new Locale("es"));
    final FormatDescriptor formatoXML
        = (FormatDescriptor) repositorio.getElementAt(0);
    assertEquals("El formato debe ser el esperado", formatoXML,
                 repositorio.getFormatos().get(0));
    assertEquals("El formato debe tener el nombre esperado",
                 "Fichero XML normal", formatoXML.getNombre());
    assertEquals("El formato debe de tener la orden de edición esperada",
                 "emacs \"%s\"", formatoXML.getPatronOrdenEdicion());

    assertTrue(
        "No debería haber orden de importación para el formato por defecto",
        formatoXML.getPatronOrdenImportacion().length()==0);
    assertEqualElems(new String[] { "xml" }, formatoXML
                     .getExtensiones());
  }

  /** Probamos que los datos se guardan bien. */
  public void testGuardado() throws FormatDescriptorException {
    final FormatDescriptor formatoXML
        = (FormatDescriptor) repositorio.getElementAt(0);
    formatoXML.setNombre("prueba");
    repositorio.guardar();

    // Comprobamos que ha cambiado el fichero
    repositorio = new FormatDescriptorRepository(
        new String[] { REPO_DIR_ALT, REPO_DIR });
    final FormatDescriptor formatoXMLb
        = (FormatDescriptor) repositorio.getElementAt(0);
    assertEquals("El formato debe tener el nombre esperado",
                 "prueba", formatoXMLb.getNombre());
    assertEquals("emacs \"%s\"", formatoXML.getPatronOrdenEdicion());
    assertTrue(
        "No debería haber orden de importación para el formato por defecto",
        formatoXML.getPatronOrdenImportacion().length()==0);
    assertEqualElems(new String[] { "xml" }, formatoXML
                     .getExtensiones());

    // Lo ponemos como estaba
    formatoXMLb.restaurar();
    repositorio.guardar();
  }

  /** Probamos que los nombres son localizables y se guardan bien. */
  public void testLocalizacion() throws FormatDescriptorException {
    final FormatDescriptor formatoXML
        = (FormatDescriptor) repositorio.getElementAt(0);

    // Activamos la local española y probamos (hacemos cambios)
    Locale.setDefault(new Locale("es"));
    assertEquals("El formato debe tener el nombre esperado",
                 "Fichero XML normal", formatoXML.getNombre());
    formatoXML.setNombre("Prueba1");

    // Activamos la local inglesa y probamos (hacemos cambios)
    Locale.setDefault(new Locale("en"));
    assertEquals("El formato debe tener el nombre esperado",
                 "Regular XML file", formatoXML.getNombre());
    formatoXML.setNombre("Prueba2");

    // Guardamos y volvemos a intentarlo
    repositorio.guardar();
    repositorio = new FormatDescriptorRepository(new String[] { REPO_DIR_ALT,
                                                                REPO_DIR });
    final FormatDescriptor formatoXMLb
        = (FormatDescriptor) repositorio.getElementAt(0);

    // Activamos la local española y probamos
    Locale.setDefault(new Locale("es"));
    assertEquals("El formato debe tener el nombre esperado",
                 "Prueba1", formatoXMLb.getNombre());

    // Activamos la local inglesa y probamos
    Locale.setDefault(new Locale("en"));
    assertEquals("El formato debe tener el nombre esperado",
                 "Prueba2", formatoXMLb.getNombre());

    formatoXML.restaurar();
    repositorio.guardar();
  }

  public void testMultiDirs() throws Exception {
    // Tomamos la fecha del fichero xml.format original
    final long modAntes = new File(REPO_DIR + File.separator + TEST_FORMAT)
        .lastModified();

    final FormatDescriptor formatoXML
        = (FormatDescriptor) repositorio.getElementAt(0);
    formatoXML.setNombre("prueba");
    repositorio.guardar();

    // El fichero original no debe haber cambiado
    final long modDespues = new File(REPO_DIR + File.separator + TEST_FORMAT)
        .lastModified();
    assertEquals("Fichero con menos prioridad no debería de haber cambiado",
                 modAntes, modDespues);

    // Los cambios han de haberse guardado en el directorio con más
    // prioridad
    final long modPrioritario = new File(REPO_DIR_ALT + File.separator
                                         + TEST_FORMAT).lastModified();
    assertTrue("Cambios deberían guardarse en el de más prioridad",
               modPrioritario >= modAntes);

    formatoXML.restaurar();
    repositorio.guardar();
  }

  public void testRestaurar() throws Exception {
    // Cambiamos y guardamos
    final FormatDescriptor formatoXML
        = (FormatDescriptor) repositorio.getElementAt(0);
    String nombreAntiguo = formatoXML.getNombre();
    formatoXML.setNombre("prueba_restaurar");
    repositorio.guardar();

    // Comprobamos que se ha cambiado, restauramos y guardamos
    repositorio = new FormatDescriptorRepository(new String[] { REPO_DIR_ALT,
                                                                REPO_DIR });
    final FormatDescriptor formatoXMLb
        = (FormatDescriptor) repositorio.getElementAt(0);
    assertEquals("prueba_restaurar", formatoXMLb.getNombre());
    formatoXMLb.restaurar();
    repositorio.guardar();

    // Volvemos a abrir y comprobamos que vuelve a estar como antes
    repositorio = new FormatDescriptorRepository(new String[] { REPO_DIR_ALT,
                                                                REPO_DIR });
    final FormatDescriptor formatoXMLc
        = (FormatDescriptor) repositorio.getElementAt(0);
    assertEquals(nombreAntiguo, formatoXMLc.getNombre());
  }

  public void testBusqueda() throws Exception {
    // Ahora pedimos por formato
    final FormatDescriptor formatoXML
        = repositorio.getFormatoPorRuta("probando.xml");
    assertMember("La búsqueda de un formato que existe funciona", "xml",
                 formatoXML.getExtensiones());

    // Comprobamos que también falla sin problemas
    final FormatDescriptor formatoFalso
        = repositorio.getFormatoPorRuta("probando.noexisto");
    assertNull("La búsqueda de un formato que no existe devuelve null",
               formatoFalso);
  }

  /* Test that proper validation is done on all fields of the format descriptor */
  public void testValidation() {
    // Get the format descriptor
    final FormatDescriptor xmlTD = repositorio.getFormatoPorRuta("probando.xml");

    try {
      xmlTD.setExtensiones(null);
      fail("Changing a format descriptor to have a null ref as its extension list should fail");
    }
    catch (FormatDescriptorException ex) {}

    try {
      xmlTD.setExtensiones(new String[]{});
      fail("Changing a format descriptor not to have any matching extensions should fail");
    }
    catch (FormatDescriptorException ex) {}

    try {
      xmlTD.setExtensiones(new String[]{"ignoreme", ""});
      fail("Changing a format descriptor to have an empty extension should fail");
    }
    catch (FormatDescriptorException ex) {}

    try {
      xmlTD.setExtensiones(new String[]{"space out"});
      fail("Changing a format descriptor to have an extension with a space should fail");
    }
    catch (FormatDescriptorException ex) {}

    try {
      xmlTD.setNombre(null);
      fail("Changing a format descriptor to have a null ref for name should fail");
    }
    catch (FormatDescriptorException ex) {}

    try {
      xmlTD.setNombre("   ");
      fail("Changing a format descriptor to have a name with only whitespace should fail");
    }
    catch (FormatDescriptorException ex) {}

    try {
      xmlTD.setPatronOrdenEdicion(null);
      fail("Changing a format descriptor to have a null ref for edit command should fail");
    }
    catch (FormatDescriptorException ex) {}

    try {
      xmlTD.setPatronOrdenEdicion("  ");
      fail("Changing a format descriptor to have an edit command with only whitespace should fail");
    }
    catch (FormatDescriptorException ex) {}

    try {
      xmlTD.setPatronOrdenEdicion(null);
      fail("Changing a format descriptor to have a null ref for import command should fail");
    }
    catch (FormatDescriptorException ex) {}

    try {
      xmlTD.setPatronOrdenImportacion("  ");
      fail("Changing a format descriptor to have a non-empty import command with only whitespace should fail");
    }
    catch (FormatDescriptorException ex) {}

    // Empty import commands should still work, as they're used for XML-based formats
    try {
      xmlTD.setPatronOrdenImportacion("");
    }
    catch (FormatDescriptorException ex) {
      fail("Changing a format descriptor to have an empty import command should succeed");
    }
  }

  protected void assertEqualElems(String[] vA, String[] vB) {
    assertEquals(vA.length, vB.length);
    for (int i = 0; i < vA.length; ++i) {
      assertEquals(vA[i], vB[i]);
    }
  }

  protected void assertMember(String msg, String aguja, String[] pajar) {
    for (String elem : pajar) {
      if (aguja.equals(elem))
        return;
    }

    fail(msg);
  }

  @Override
  public void setUp() throws FormatDescriptorException {
    repositorio = new FormatDescriptorRepository(new String[] { REPO_DIR_ALT,
                                                                REPO_DIR });
  }

  @Override
  public void tearDown() throws FormatDescriptorException {
    for (FormatDescriptor fd : repositorio.getFormatos()) {
      fd.restaurar();
    }
    repositorio = null;
  }

}